#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------
# Copyright 2021 CFS Engineering
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------

# Author: Aidan Jungo

import math

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from cpacspy.cpacsfunctions import (add_float_vector, create_branch,
                                    get_float_vector, get_xpath_parent)
from cpacspy.utils import (AEROPERFORMANCE_XPATH, COEFS, PARAMS, PARAMS_COEFS, listify)



def get_filter(df,list_of,alt_list,mach_list,aos_list,aoa_list):
    """ Get a dataframe filter for a set of parameters lists. """

    filt = df[list_of] ==  df[list_of]

    if alt_list:
        filt &= df['altitude'].isin(alt_list)

    if mach_list:
        filt &= df['machNumber'].isin(mach_list)

    if aos_list:
        filt &= df['angleOfSideslip'].isin(aos_list)

    if aoa_list:
        filt &= df['angleOfAttack'].isin(aoa_list)

    return filt


class AeroMap:
    """ AeroMap class for CPACS AeroMap. """
    
    def __init__(self, tixi, uid, create_new=False):
        """ Init aeromap class

        Args:
            tixi (object): TIXI object open from a CPACS file
            uid (str): UID of the AeroMap
            create_new (bool, optional): If True crate a new AeroMap in TIXI, if false find it in the CPACS file. Defaults to False.
        """
        
        self.tixi = tixi
        self.uid = uid
        self.name = uid
        self.description = ''
        self.df = pd.DataFrame(columns=PARAMS_COEFS)

        if create_new:
            self.name = uid
            self.xpath = None 

        else:
            self.xpath = self.tixi.uIDGetXPath(uid) + '/aeroPerformanceMap'
            xpath_name = get_xpath_parent(self.xpath) + '/name'
            if self.tixi.checkElement(xpath_name):
                self.name = self.tixi.getTextElement(xpath_name)
            
            xpath_description = get_xpath_parent(self.xpath) + '/description'
            if self.tixi.checkElement(xpath_description):
                self.description = self.tixi.getTextElement(xpath_description)

            self.get_param_and_coef()

    def get_param_and_coef(self):
        """ Get the parameters and coefficients from the CPACS file."""

        param_dict = {}

        for param in PARAMS_COEFS:
            param_xpath = self.xpath + f'/{param}'

            if self.tixi.checkElement(param_xpath):
                param_dict[param] = get_float_vector(self.tixi,param_xpath)

            else:
                if param in PARAMS:
                    raise ValueError(f'No values has been found for "{param}" in "{self.uid}" aeroMap!')
        
        df_param  =  pd.DataFrame(param_dict)
        self.df = pd.concat([self.df, df_param], axis=0)

    def get(self,list_of,alt=None,mach=None,aos=None,aoa=None):
        """ Get parameter or coefficient as a numpy vector with other parameters as filter (optional)."""

        alt_list = listify(alt)
        mach_list = listify(mach)
        aos_list = listify(aos)
        aoa_list = listify(aoa)

        filt = get_filter(self.df,list_of,alt_list,mach_list,aos_list,aoa_list)
        
        return self.df.loc[filt,list_of].to_numpy()

    def add_values(self,alt,mach,aos,aoa,cd=np.nan,cl=np.nan,cs=np.nan,cmd=np.nan,cml=np.nan,cms=np.nan): 
        """ Add a row in an Aeromap dataframe."""

        new_row = {'altitude': alt, 'machNumber': mach, 'angleOfSideslip': aos, 'angleOfAttack': aoa,
                   'cd': cd, 'cl': cl, 'cs': cs, 'cmd': cmd, 'cml': cml, 'cms': cms}

        self.df = self.df.append(new_row, ignore_index=True)

    def plot(self,x_param,y_param,alt=None,mach=None,aos=None,aoa=None):
        """ Plot 'x_param' vs 'y_param' with filtered parameters passed as float or string. """

        alt_list = listify(alt)
        mach_list = listify(mach)
        aos_list = listify(aos)
        aoa_list = listify(aoa)

        filt = get_filter(self.df,x_param,alt_list,mach_list,aos_list,aoa_list)
        self.df.loc[filt].plot(x=x_param,y=y_param,marker='o')
        plt.show()

    def save(self):
        """ Save the AeroMap in the TIXI object. """

        # Create and fill the '/aeroPerformanceMap' field
        if not self.xpath:
            if self.tixi.checkElement(AEROPERFORMANCE_XPATH):
                child_nb = self.tixi.getNumberOfChilds(AEROPERFORMANCE_XPATH)
                self.tixi.createElementAtIndex(AEROPERFORMANCE_XPATH,'aeroMap',child_nb+1)
                self.xpath = AEROPERFORMANCE_XPATH + f'/aeroMap[{child_nb+1}]/aeroPerformanceMap'
            else:
                create_branch(self.tixi,AEROPERFORMANCE_XPATH+'/aeroMap')
                self.xpath = AEROPERFORMANCE_XPATH + '/aeroMap/aeroPerformanceMap'

            self.tixi.uIDSetToXPath(get_xpath_parent(self.xpath), self.uid)
            create_branch(self.tixi,self.xpath)

        # Create and fill parameters fields
        for param in PARAMS:
            if param in self.df:
                if not self.df[param].isnull().values.any():
                    param_xpath = self.xpath + '/' + param
                    create_branch(self.tixi,param_xpath)
                    add_float_vector(self.tixi,param_xpath,self.df[param].tolist())
                else:
                    raise ValueError('All the 4 parametres (alt,mach,aos,aoa) must be define to save an aeroMap!')
            else:
                raise ValueError('All the 4 parametres (alt,mach,aos,aoa) must be define to save an aeroMap!')

        # Create and fill coefficients fields
        for coef in COEFS:
            if coef in self.df:
                if not self.df[coef].isnull().values.any():
                    coef_xpath = self.xpath + '/' + coef
                    create_branch(self.tixi,coef_xpath)
                    add_float_vector(self.tixi,coef_xpath,self.df[coef].tolist())
            
        # Create and fill the '/name' field
        name_xpath = get_xpath_parent(self.xpath) + '/name'
        create_branch(self.tixi,name_xpath)
        self.tixi.updateTextElement(name_xpath,self.uid)
        
        # Create and fill the '/description' field
        description_xpath = get_xpath_parent(self.xpath) + '/description'
        create_branch(self.tixi,description_xpath)
        self.tixi.updateTextElement(description_xpath,self.description)

        # Create and fill the '/atmosphericModel' field
        atm_model_xpath = get_xpath_parent(self.xpath) + '/boundaryConditions/atmosphericModel'
        if not self.tixi.checkElement(atm_model_xpath):
            create_branch(self.tixi,atm_model_xpath)
        self.tixi.updateTextElement(atm_model_xpath,'ISA')

    def export_csv(self, csv_path):
        """ Export the AeroMap as a CSV file. """

        self.df.to_csv(csv_path, na_rep='NaN', index=False)

    def get_cd0_oswald(self,ar,alt=None,mach=None,aos=None,plot=False):
        """ Calculate and return CD0 and Oswald factor. """

        # Check for unique angleOfAttack condition
        aoa = self.get('angleOfAttack',alt,mach,aos)
        aoa_unique = list(set(aoa))

        if len(aoa) > len(aoa_unique):
            raise ValueError('You must have unique angle of attack value to calculate CD0 and Oswald factor!')

        # Get coefficient
        cd = self.get('cd',alt,mach,aos)
        cl = self.get('cl',alt,mach,aos)
        cl2 = np.power(cl, 2)
        
        # Remove value when Cl < 0
        cond = cl >= 0
        cl2 = cl2[cond]
        cd = cd[cond]
        cl = cl[cond]
        
        # Linear regression
        coef = np.polyfit(cl2, cd, 1)
        f = np.poly1d(coef)
        x = np.arange(0,2,0.01)
        y = f(x)

        # Calculate CD0 and Oswald factor e
        k, cd0 = coef
        e = 1 / (k*ar*math.pi)

        print('---------------------------------------------')
        print(f'For alt={alt}m, Mach={mach}, AoS={aos}deg:')
        print('CD0:',round(cd0,6))
        print('Oswald factor:',round(e,4))
        print('---------------------------------------------')

        if plot:
            fig, ax = plt.subplots()
            ax.plot(cl2, cd,'o')
            ax.plot(x, y,'-')
            plt.show()

        return cd0,e

    def get_forces(self,aircraft):
        """ Get forces """
        # TODO
        pass

    def __str__(self): 

        text_line = []
        text_line.append('\nAeroMap Description --------------------------------------------------------------------------------')
        text_line.append(f'\nAeroMap uid: {self.uid}')
        text_line.append(f'AeroMap xpath: {self.xpath}')
        text_line.append(f'AeroMap description: {self.description}')
        text_line.append(' ')
        text_line.append(f'Number of states: \t\t\t{self.df["angleOfAttack"].size}')
        text_line.append(f'Unique altitude: \t\t\t{str(self.df["altitude"].unique().tolist()).strip("[]")}')
        text_line.append(f'Unique Mach number: \t\t{str(self.df["machNumber"].unique().tolist()).strip("[]")}')
        text_line.append(f'Unique angle of sideslip:  \t{str(self.df["angleOfSideslip"].unique().tolist()).strip("[]")}')
        text_line.append(f'Unique angle of attack: \t{str(self.df["angleOfAttack"].unique().tolist()).strip("[]")}')
        text_line.append(' ')
        text_line.append('Parameters and coefficients:')
        text_line.append(str(self.df))
        text_line.append('----------------------------------------------------------------------------------------------------\n')

        return ('\n').join(text_line)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------
# Copyright 2021 CFS Engineering
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------

# Author: Aidan Jungo

from cpacspy.cpacsfunctions import (get_value_or_default, open_tigl, get_tigl_aircraft)
from cpacspy.utils import REF_XPATH


class Aircraft:

    def __init__(self,tixi):
        """ Aircraft class to store references values and other information about the aircraft

        Args:
            tixi (object): TIXI object open from a CPACS file
        """
        
        self.tixi = tixi
        self.tigl = open_tigl(self.tixi)

        # Reference values
        self.ref_lenght = get_value_or_default(self.tixi,REF_XPATH + '/length',1)
        self.ref_area = get_value_or_default(self.tixi,REF_XPATH + '/area',1)
        self.ref_point_x = get_value_or_default(self.tixi,REF_XPATH + '/point/x',0)
        self.ref_point_y = get_value_or_default(self.tixi,REF_XPATH + '/point/y',0)
        self.ref_point_z = get_value_or_default(self.tixi,REF_XPATH + '/point/z',0)

        # Aircraft specific values (extract with TiGL)
        self.aircraft_tigl = get_tigl_aircraft(self.tigl)
        self.ref_wing_idx = 1 # By defauld the reference wing is "1", could be changed by updating this attribute
  
    # When self.ref_wing_idx is change:  
    # TODO: change it by uid also
    @property
    def ref_wing_idx(self):
        return self._ref_wing_idx

    @ref_wing_idx.setter
    def ref_wing_idx(self, new_idx):
        self._ref_wing_idx = new_idx

        self.wing_half_span = self.aircraft_tigl.get_wing(self._ref_wing_idx).get_wing_half_span()
        self.wing_area = self.aircraft_tigl.get_wing(self._ref_wing_idx).get_surface_area()
        self.wing_ar = self.aircraft_tigl.get_wing(self._ref_wing_idx).get_aspect_ratio()
        
    def __str__(self): 

        text_line = []
        text_line.append('\nAircraft data --------------------------------------------------------------------------------------')
        text_line.append(' ')
        text_line.append(f'Reference lengh: \t{self.ref_lenght} [m]')
        text_line.append(f'Reference area: \t{self.ref_area} [m^2]')
        text_line.append(f'Reference point: \t({self.ref_point_x},{self.ref_point_y},{self.ref_point_z}) [m]')
        text_line.append(' ')
        text_line.append(f'Reference wing index: \t{self._ref_wing_idx}')
        text_line.append(f'Wing half span: \t{self.wing_half_span} [m]')
        text_line.append(f'Wing area: \t\t{self.wing_area} [m^2]')
        text_line.append(f'Wing AR: \t\t{self.wing_ar} [-]')
        text_line.append(' ')
        text_line.append('----------------------------------------------------------------------------------------------------\n')
        return ('\n').join(text_line)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------
# Copyright 2021 CFS Engineering
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------

# Author: Aidan Jungo

import tigl3.configuration
import tixi3.tixi3wrapper as tixi3wrapper
import tigl3.tigl3wrapper as tigl3wrapper
from tixi3.tixi3wrapper import Tixi3Exception
from tigl3.tigl3wrapper import Tigl3Exception


def open_tixi(cpacs_path):
    """ Create the TIXI Handle of a CPACS file given as input
    by its path. If this operation is not possible, it returns 'None'

    Source :
        * TIXI functions: http://tixi.sourceforge.net/Doc/index.html

    Args:
        cpacs_path (str): Path to the CPACS file

    Returns::
        tixi_handle (handles): TIXI Handle of the CPACS file
    """

    tixi_handle = tixi3wrapper.Tixi3()
    tixi_handle.open(cpacs_path)

    print(f'TIXI handle has been created for {cpacs_path}.')

    return tixi_handle


def open_tigl(tixi_handle):
    """ Function 'open_tigl' return the TIGL Handle from its TIXI Handle.
    If this operation is not possible, it returns 'None'

    Source :
        * TIGL functions http://tigl.sourceforge.net/Doc/index.html

    Args:
        tixi_handle (handles): TIXI Handle of the CPACS file

    Returns:
        tigl_handle (handles): TIGL Handle of the CPACS file
    """

    # Get model uid, requierd to open TiGL handle in case there is also a rotorcraft in the CPACS file
    model_xpath = '/cpacs/vehicles/aircraft/model'
    if tixi_handle.checkAttribute(model_xpath, 'uID'):
        model_uid = tixi_handle.getTextAttribute(model_xpath, 'uID')
    else:
        # log.warning('No model uID in the CPACS file!')
        model_uid = ''

    tigl_handle = tigl3wrapper.Tigl3()
    tigl_handle.open(tixi_handle, model_uid)

    tigl_handle.logSetVerbosity(1)  # 1 - only error, 2 - error and warnings

    # log.info('TIGL handle has been created.')
    return tigl_handle


def get_tigl_aircraft(tigl):
    """ Get the TiGL aircraft configuration manager. """
    
    # Get the configuration manager
    mgr =  tigl3.configuration.CCPACSConfigurationManager_get_instance()
    aircraft = mgr.get_configuration(tigl._handle.value)
    
    return aircraft


def get_value(tixi, xpath):
    """ Check first if the the xpath exist and a value is store
    at this place. Then, it gets and returns this value. If the value or the
    xpath does not exist it raise an error and return 'None'.

    Source :
        * TIXI functions: http://tixi.sourceforge.net/Doc/index.html

    Args:
        tixi_handle (handles): TIXI Handle of the CPACS file
        xpath (str): xpath of the value to get

    Returns:
         value (float or str): Value found at xpath
    """

    # Try to get the a value at xpath
    try:
        value = tixi.getTextElement(xpath)
    except:
        value = None

    if value:
        try: # check if it is a 'float'
            is_float = isinstance(float(value), float)
            value = float(value)
        except:
            pass
    else:
        # check if the path exist
        if tixi.checkElement(xpath):
            # log.error('No value has been found at ' + xpath)
            raise ValueError('No value has been found at ' + xpath)
        else:
            # log.error(xpath + ' cannot be found in the CPACS file')
            raise ValueError(xpath + ' cannot be found in the CPACS file')

    # Special return for boolean
    if value == 'True':
        return True
    elif value == 'False':
        return False

    return value


def get_value_or_default(tixi,xpath,default_value):
    """ Do the same than the function 'get_value'
    but if no value is found at this place it returns the default value and add
    it at the xpath. If the xpath does not exist, it is created.

    Source :
        * TIXI functions: http://tixi.sourceforge.net/Doc/index.html

    Args:
        tixi_handle (handles): TIXI Handle of the CPACS file
        xpath (str): xpath of the value to get
        default_value (str, float or int): Default value

    Returns:
        tixi (handles): Modified TIXI Handle (with added default value)
        value (str, float or int): Value found at xpath
    """

    value = None
    try:
        value = get_value(tixi, xpath)
    except:
        pass

    if value is None:
        # log.info('Default value will be used instead')
        value = default_value

        xpath_parent = '/'.join(str(m) for m in xpath.split("/")[:-1])
        value_name = xpath.split("/")[-1]
        create_branch(tixi,xpath_parent,False)

        is_int = False
        is_float = False
        is_bool = False
        try: # check if it is an 'int' or 'float'
            is_int = isinstance(float(default_value), int)
            is_float = isinstance(float(default_value), float)
            is_bool = isinstance(default_value, bool)
        except:
            pass
        if is_bool:
           tixi.addTextElement(xpath_parent,value_name,str(value))
        elif is_float or is_int:
            value = float(default_value)
            tixi.addDoubleElement(xpath_parent,value_name,value,'%g')
        else:
            value = str(value)
            tixi.addTextElement(xpath_parent,value_name,value)
    else:
        # Special return for boolean
        if value == 'True':
            return True
        elif value == 'False':
            return False
        elif isinstance(value,bool):
            return value

    return value


def get_float_vector(tixi, xpath):
    """ Get a vector (composed by float) at the
    given XPath, if the node does not exist, an error will be raised.

    Args:
        tixi (handle): Tixi handle
        xpath (str): XPath of the vector to get
    """

    if not tixi.checkElement(xpath):
        raise ValueError(xpath + ' path does not exist!')

    float_vector_str = tixi.getTextElement(xpath)

    if float_vector_str == '':
        raise ValueError('No value has been fournd at ' + xpath)

    if float_vector_str.endswith(';'):
        float_vector_str = float_vector_str[:-1]
    float_vector_list = float_vector_str.split(';')
    float_vector = [float(elem) for elem in float_vector_list]

    return float_vector


def add_float_vector(tixi, xpath, vector):
    """ Add a vector (composed by float) at the given XPath, 
    if the node does not exist, it will be created. Values will be
    overwritten if paths exists.

    Args:
        tixi (handle): Tixi handle
        xpath (str): XPath of the vector to add
        vector (list, tuple): Vector of floats to add
    """

    # Strip trailing '/' (has no meaning here)
    if xpath.endswith('/'):
        xpath = xpath[:-1]

    # Get the field name and the parent CPACS path
    xpath_child_name = xpath.split("/")[-1]
    xpath_parent = xpath[:-(len(xpath_child_name)+1)]

    if not tixi.checkElement(xpath_parent):
        create_branch(tixi,xpath_parent)

    vector = [float(v) for v in vector]

    if tixi.checkElement(xpath):
        tixi.updateFloatVector(xpath, vector, len(vector), format='%g')
        tixi.addTextAttribute(xpath, 'mapType', 'vector')
    else:
        tixi.addFloatVector(xpath_parent, xpath_child_name, vector, \
                            len(vector), format='%g')
        tixi.addTextAttribute(xpath, 'mapType', 'vector')


def get_xpath_parent(xpath,level=1):
    """ Get the parent xpath at any level, 1 is parent just above the input xpath.

    Args:
        xpath (str): Input xpath
        level (int, optional): Parent level. Defaults to 1.

    Returns:
        str: Parent xpath
    """

    if not xpath.startswith('/'):
        raise ValueError('"get_xpath_parent" must recieve an xpath as argument!')

    if len(xpath.split('/'))-1 <= level:
        raise ValueError('No parent available at this level')

    return '/'.join(xpath.split('/')[:-level])


def create_branch(tixi, xpath, add_child=False):
    """ Create a branch in the tixi handle and also all the missing parent nodes. 
    Be careful, the xpath must be unique until the last element, it means, 
    if several element exist, its index must be precised (index start at 1).
    e.g.: '/cpacs/vehicles/aircraft/model/wings/wing[2]/name'

    If the entire xpath already exist, the option 'add_child' (True/False) lets
    the user decide if a named child should be added next to the existing
    one(s). This only valid for the last element of the xpath.

    Source :
        * TIXI functions: http://tixi.sourceforge.net/Doc/index.html

    Args:
        tixi (handles): TIXI Handle of the CPACS file
        xpath (str): xpath of the branch to create
        add_child (boolean): Choice of adding a name child if the last element
                             of the xpath if one already exists

    Returns:
        tixi (handles): Modified TIXI Handle (with new branch)
    """

    xpath_split = xpath.split("/")
    xpath_count = len(xpath_split)

    for i in range(xpath_count-1):
        xpath_index = i + 2
        xpath_partial = '/'.join(str(m) for m in xpath_split[0:xpath_index])
        xpath_parent = '/'.join(str(m) for m in xpath_split[0:xpath_index-1])
        child = xpath_split[(xpath_index-1)]
        if tixi.checkElement(xpath_partial):
            if child == xpath_split[-1] and add_child:
                namedchild_nb = tixi.getNamedChildrenCount(xpath_parent, child)
                tixi.createElementAtIndex (xpath_parent,child,namedchild_nb+1)
        else:
            tixi.createElement(xpath_parent, child)
            


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import setuptools
import os

NAME = 'cpacspy'
VERSION = '0.0.1'
AUTHOR = 'Aidan Jungo'
EMAIL = 'aidan.jungo@cfse.ch'
DESCRIPTION = 'PyPI package to work with CPACS file and AeroMaps'
LONG_DESCRIPTION = 'Python package to read, write and analyse CPACS AeroPerformanceMaps. You need to have TIXI and TIGL install on your computer to use this package.'
URL = ''
REQUIRES_PYTHON = '>=3.6.0'
REQUIRED = []
README = 'README.rst'
PACKAGE_DIR = 'src/'
LICENSE = 'Apache License 2.0'



setuptools.setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    author_email=EMAIL,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    url=URL,
    include_package_data=True,
    package_dir={'': PACKAGE_DIR},
    license=LICENSE,
    packages=[NAME],
    python_requires=REQUIRES_PYTHON,
    keywords=['CPACS','aircraft','design','xml','aerodynamics','coefficients'],
    install_requires=REQUIRED,
    # See: https://pypi.org/classifiers/
    classifiers=[
        "Programming Language :: Python :: 3",
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
    ],
)
